@@ -577,6 +577,7 @@ CHAR_P perl_cmd_module (cmd_parms *parms, void *dummy, char *arg)
     if(PERL_RUNNING()) {
 	if (PERL_STARTUP_IS_DONE) {
 	    if (perl_require_module(arg, NULL) != OK) {
+		dTHR;
 		STRLEN n_a;
 		return SvPV(ERRSV,n_a);
 	    }
@@ -605,6 +606,7 @@ CHAR_P perl_cmd_require (cmd_parms *parms, void *dummy, char *arg)
     if(PERL_RUNNING()) {
 	if (PERL_STARTUP_IS_DONE) {
 	    if (perl_load_startup_script(NULL, parms->pool, arg, TRUE) != OK) {
+		dTHR;
 		STRLEN n_a;
 		return SvPV(ERRSV,n_a);
 	    }
@@ -1127,6 +1129,7 @@ void perl_section_hash_walk(cmd_parms *cmd, void *cfg, HV *hv)
     while ((tmpval = hv_iternextsv(hv, &tmpkey, &tmpklen))) { 
 	char line[MAX_STRING_LEN]; 
 	char *value = NULL;
+	if (SvMAGICAL(tmpval)) mg_get(tmpval); /* tied hash FETCH */
 	if(SvROK(tmpval)) {
 	    if(SvTYPE(SvRV(tmpval)) == SVt_PVAV) {
 		perl_handle_command_av((AV*)SvRV(tmpval), 
@@ -1202,6 +1205,11 @@ CHAR_P perl_virtualhost_section (cmd_parms *cmd, void *dummy, HV *hv)
     main_server->next = s;
     cmd->server = s;
 
+#if MODULE_MAGIC_AT_LEAST(19990320, 5)
+    s->defn_name = cmd->config_file->name;
+    s->defn_line_number = cmd->config_file->line_number;
+#endif
+
     TRACE_SECTION("VirtualHost", arg);
 
     perl_section_hash_walk(cmd, s->lookup_defaults, tab);
@@ -1435,15 +1443,25 @@ CHAR_P perl_end_section (cmd_parms *cmd, void *dummy) {
     return perl_end_magic;
 }
 
+#define STRICT_PERL_SECTIONS_SV \
+perl_get_sv("Apache::Server::StrictPerlSections", FALSE)
+
 void perl_handle_command(cmd_parms *cmd, void *config, char *line) 
 {
     CHAR_P errmsg;
+    SV *sv;
 
     MP_TRACE_s(fprintf(stderr, "handle_command (%s): ", line));
-    errmsg = handle_command(cmd, config, line);
+    if ((errmsg = handle_command(cmd, config, line))) {
+	if ((sv = STRICT_PERL_SECTIONS_SV) && SvTRUE(sv)) {
+	    croak("<Perl>: %s", errmsg);
+	}
+	else {
+	    log_printf(cmd->server, "<Perl>: %s", errmsg);
+	}
+    }
+
     MP_TRACE_s(fprintf(stderr, "%s\n", errmsg ? errmsg : "OK"));
-    if(errmsg)
-	log_printf(cmd->server, "<Perl>: %s", errmsg);
 }
 
 void perl_handle_command_hv(HV *hv, char *key, cmd_parms *cmd, void *config)
